# -*- coding: utf-8 -*-

import re
from .regexp import Regexp, RegexpString
from .base import String, FromBytes, OnError, WithRepr
from .lib import py3
from . import codes

if py3:
    import urllib.parse as urlparse
else:  # pragma: no cover
    import urlparse


MAX_EMAIL_LEN = 254


EMAIL_REGEXP = re.compile(
    # dot-atom
    r"(?P<name>^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
    # quoted-string
    r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"'
    # domain
    r')@(?P<domain>(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)$)'
    # literal form, ipv4 address (SMTP 4.1.3)
    r'|\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$',
    re.IGNORECASE,
)


def email_idna_encode(value):
    if '@' in value:
        parts = value.split('@')
        parts[-1] = parts[-1].encode('idna').decode('ascii')
        return '@'.join(parts)
    return value


to_str = OnError(FromBytes('utf-8') | String(), 'value is not a string', code=codes.IS_NOT_A_STRING)


email_regexp_trafaret = OnError(to_str & Regexp(EMAIL_REGEXP), 'value is not a valid email address')
email_trafaret = (email_regexp_trafaret | (to_str & email_idna_encode & email_regexp_trafaret))
Email = to_str & OnError(
    String(max_length=MAX_EMAIL_LEN) & email_trafaret,
    'value is not a valid email address',
    code=codes.IS_NOT_VALID_EMAIL,
)
Email = WithRepr(Email, '<Email>')


class Hex(RegexpString):
    regex = r'^[0-9a-f]*$'
    str_method = "lower"

    def __repr__(self):
        return '<Hex>'


class URLSafe(RegexpString):
    regex = r'^[0-9A-Za-z-_]*$'

    def __repr__(self):
        return '<URLSafe>'


URL_REGEXP = re.compile(
    r'^(?:http|ftp)s?://'  # http:// or https://
    r'(?:\S+(?::\S*)?@)?'  # user and password
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-_]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$',
    re.IGNORECASE,
)
URLRegexp = to_str & Regexp(URL_REGEXP)


def decode_url_idna(value):
    scheme, netloc, path, query, fragment = urlparse.urlsplit(value)
    netloc = netloc.encode('idna').decode('ascii')  # IDN -> ACE
    return urlparse.urlunsplit((scheme, netloc, path, query, fragment))


URL = OnError(
    URLRegexp | (to_str & decode_url_idna & URLRegexp),
    'value is not URL',
    code=codes.IS_NOT_VALID_URL,
)
URL = WithRepr(URL, '<URL>')


IPv4 = OnError(
    Regexp(
        r'^((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])$',  # noqa
    ),
    'value is not IPv4 address',
    code=codes.IS_NOT_IPv4,
)
IPv4 = WithRepr(IPv4, '<IPv4>')


IPv6 = OnError(
    Regexp(
        r'^('
        r'(::)|'
        r'(::[0-9a-f]{1,4})|'
        r'([0-9a-f]{1,4}:){7,7}[0-9a-f]{1,4}|'
        r'([0-9a-f]{1,4}:){1,7}:|'
        r'([0-9a-f]{1,4}:){1,6}:[0-9a-f]{1,4}|'
        r'([0-9a-f]{1,4}:){1,5}(:[0-9a-f]{1,4}){1,2}|'
        r'([0-9a-f]{1,4}:){1,4}(:[0-9a-f]{1,4}){1,3}|'
        r'([0-9a-f]{1,4}:){1,3}(:[0-9a-f]{1,4}){1,4}|'
        r'([0-9a-f]{1,4}:){1,2}(:[0-9a-f]{1,4}){1,5}|'
        r'[0-9a-f]{1,4}:((:[0-9a-f]{1,4}){1,6})|'
        r':((:[0-9a-f]{1,4}){1,7}:)|'
        r'fe80:(:[0-9a-f]{0,4}){0,4}%[0-9a-z]{1,}|'
        r'::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|'  # noqa
        r'([0-9a-f]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])'  # noqa
        r')$',
        re.IGNORECASE,
    ),
    'value is not IPv6 address',
    code=codes.IS_NOT_IPv6,
)
IPv6 = WithRepr(IPv6, '<IPv6>')


IP = OnError(IPv4 | IPv6, 'value is not IP address', code=codes.IS_NOT_IP)
IP = WithRepr(IP, '<IP>')



print("import: 'trafaret'")
import trafaret

print("import: 'trafaret.constructor'")
import trafaret.constructor

print("import: 'trafaret.keys'")
import trafaret.keys

print("import: 'trafaret.utils'")
import trafaret.utils

print("import: 'trafaret.contrib.rfc_3339'")
import trafaret.contrib.rfc_3339




# -*- coding: utf-8 -*-

import functools
import itertools
import warnings
from datetime import date, datetime
from .lib import (
    py3,
    py36,
    py3metafix,
    getargspec,
    get_callable_args,
    with_context_caller,
    _empty,
    STR_TYPES,
    AbcMapping,
    Iterable as AbcIterable,
)
from .dataerror import DataError
from . import codes


if py36:
    from .async_mixins import (
        TrafaretAsyncMixin,
        OrAsyncMixin,
        AndAsyncMixin,
        ListAsyncMixin,
        TupleAsyncMixin,
        MappingAsyncMixin,
        CallAsyncMixin,
        ForwardAsyncMixin,
        DictAsyncMixin,
        KeyAsyncMixin,
    )
else:  # pragma: no cover
    class EmptyMixin(object):
        pass
    TrafaretAsyncMixin = EmptyMixin
    OrAsyncMixin = EmptyMixin
    AndAsyncMixin = EmptyMixin
    ListAsyncMixin = EmptyMixin
    TupleAsyncMixin = EmptyMixin
    MappingAsyncMixin = EmptyMixin
    CallAsyncMixin = EmptyMixin
    ForwardAsyncMixin = EmptyMixin
    DictAsyncMixin = EmptyMixin
    KeyAsyncMixin = EmptyMixin


# Python3 support
if py3:
    unicode = str
    BYTES_TYPE = bytes
    STR_TYPE = str
else:  # pragma: no cover
    BYTES_TYPE = str
    STR_TYPE = unicode


def deprecated(message):
    warnings.warn(message, DeprecationWarning)


"""
Trafaret is tiny library for data validation
It provides several primitives to validate complex data structures
Look at doctests for usage examples
"""


class TrafaretMeta(type):
    """
    Metaclass for trafarets to make using "|" operator possible not only
    on instances but on classes

    >>> Int | String
    <Or(<Int>, <String>)>
    >>> Int | String | Null
    <Or(<Int>, <String>, <Null>)>
    >>> (Int >> (lambda v: v if v ** 2 > 15 else 0)).check(5)
    5
    """

    def __or__(cls, other):
        return cls() | other

    def __and__(cls, other):
        return cls() & other

    def __rshift__(cls, other):
        return cls() & other


@py3metafix
class Trafaret(TrafaretAsyncMixin):
    """
    Base class for trafarets, provides only one method for
    trafaret validation failure reporting
    """

    __metaclass__ = TrafaretMeta

    def check(self, value, context=None):
        """
        Common logic. In subclasses you need to implement check_value or
        check_and_return.
        """
        if hasattr(self, 'transform'):
            return self.transform(value, context=context)
        elif hasattr(self, 'check_value'):
            self.check_value(value)
            return value
        elif hasattr(self, 'check_and_return'):
            return self.check_and_return(value)
        else:
            cls = "{}.{}".format(
                type(self).__module__,
                type(self).__name__,
            )
            raise NotImplementedError(
                "You must implement check_value or"
                " check_and_return methods '%s'" % cls
            )

    def is_valid(self, value):
        """
        Allows to check value and get bool, is value valid or not.
        """
        try:
            self.check(value)
            return True
        except DataError:
            return False

    def _failure(self, error=None, value=_empty, code=None):
        """
        Shortcut method for raising validation error
        """
        raise DataError(error=error, value=value, trafaret=self, code=code)

    def __or__(self, other):
        return Or(self, other)

    def __and__(self, other):
        return And(self, other)

    def __rshift__(self, other):
        return And(self, other)

    def __call__(self, val, context=None):
        return self.check(val, context=context)


class OnError(Trafaret):
    def __init__(self, trafaret, message, code=None):
        self.trafaret = ensure_trafaret(trafaret)
        self.message = message
        self.code = code

    def transform(self, value, context=None):
        try:
            return self.trafaret(value, context=context)
        except DataError as de:
            raise DataError(
                self.message,
                value=value,
                trafaret=self.trafaret,
                code=self.code or de.code,
            )


class WithRepr(Trafaret):
    def __init__(self, trafaret, representation):
        self.trafaret = ensure_trafaret(trafaret)
        self.representation = representation

    def transform(self, value, context=None):
        return self.trafaret(value, context=context)

    def __repr__(self):
        return self.representation


def ensure_trafaret(trafaret):
    """
    Helper for complex trafarets, takes trafaret instance or class
    and returns trafaret instance
    """
    if isinstance(trafaret, Trafaret):
        return trafaret
    elif isinstance(trafaret, type):
        if issubclass(trafaret, Trafaret):
            return trafaret()
        # str, int, float are classes, but its appropriate to use them
        # as trafaret functions
        return Call(lambda val: trafaret(val))
    elif callable(trafaret):
        return Call(trafaret)
    else:
        raise RuntimeError("%r should be instance or subclass"
                           " of Trafaret" % trafaret)


class TypeMeta(TrafaretMeta):
    def __getitem__(self, type_):
        return self(type_)


@py3metafix
class TypingTrafaret(Trafaret):
    """A trafaret used for instance type and class inheritance checks."""

    __metaclass__ = TypeMeta

    def __init__(self, type_):
        self.type_ = type_

    def check_value(self, value):
        if not self.typing_checker(value, self.type_):
            self._failure(
                self.failure_message % self.type_.__name__,
                value=value,
                code=self.code,
            )

    def __repr__(self):
        return "<%s(%s)>" % (self.__class__.__name__, self.type_.__name__)


class Subclass(TypingTrafaret):
    """
    >>> Subclass(type)
    <Subclass(type)>
    >>> Subclass[type]
    <Subclass(type)>
    >>> s = Subclass[type]
    >>> s.check(type)
    <type 'type'>
    >>> extract_error(s, object)
    'value is not subclass of type'
    """

    typing_checker = issubclass
    failure_message = "value is not subclass of %s"
    code = "is_not_subclass"


class Type(TypingTrafaret):
    """
    >>> Type(int)
    <Type(int)>
    >>> Type[int]
    <Type(int)>
    >>> c = Type[int]
    >>> c.check(1)
    1
    >>> extract_error(c, "foo")
    'value is not int'
    >>> c.is_valid("foo")
    False
    """

    typing_checker = isinstance
    failure_message = "value is not %s"
    code = "is_not_instance"


class Any(Trafaret):
    """
    >>> Any()
    <Any>
    >>> (Any() >> ignore).check(object())
    """

    def check_value(self, value):
        pass

    def __repr__(self):
        return "<Any>"


@py3metafix
class Or(Trafaret, OrAsyncMixin):
    """
    >>> nullString = Or(String, Null)
    >>> nullString
    <Or(<String>, <Null>)>
    >>> nullString.check(None)
    >>> nullString.check("test")
    'test'
    >>> extract_error(nullString, 1)
    {0: 'value is not a string', 1: 'value should be None'}
    >>> nullString.is_valid(None)
    True
    >>> nullString.is_valid("test")
    True
    >>> nullString.is_valid(1)
    False
    """

    __slots__ = ['trafarets']

    def __init__(self, *trafarets):
        self.trafarets = [ensure_trafaret(t) for t in trafarets]

    def transform(self, value, context=None):
        errors = []
        for trafaret in self.trafarets:
            try:
                return trafaret(value, context=context)
            except DataError as e:
                errors.append(e)
        raise self._failure(dict(enumerate(errors)), code=codes.NOTHING_MATCH)

    def __repr__(self):
        return "<Or(%s)>" % (", ".join(repr(t) for t in self.trafarets))


class And(Trafaret, AndAsyncMixin):
    """
    Will work over trafarets sequentially
    """
    __slots__ = ('trafaret', 'other', 'disable_old_check_convert')

    def __init__(self, trafaret, other):
        self.trafaret = ensure_trafaret(trafaret)
        self.other = ensure_trafaret(other)

    def transform(self, value, context=None):
        # it will raise in case of error
        res = self.trafaret(value, context=context)
        return self.other(res, context=context)

    def __repr__(self):
        return "<And(%s, %s)>" % (
            repr(self.trafaret),
            repr(self.other),
        )


class Null(Trafaret):
    """
    >>> Null()
    <Null>
    >>> Null().check(None)
    >>> extract_error(Null(), 1)
    'value should be None'
    >>> Null().is_valid(None)
    True
    >>> Null().is_valid(1)
    False
    """

    def check_value(self, value):
        if value is not None:
            self._failure("value should be None", value=value, code=codes.IS_NOT_NULL)

    def __repr__(self):
        return "<Null>"


class Bool(Trafaret):
    """
    >>> Bool()
    <Bool>
    >>> Bool().check(True)
    True
    >>> Bool().check(False)
    False
    >>> extract_error(Bool(), 1)
    'value should be True or False'
    >>> Null().is_valid(True)
    True
    >>> Null().is_valid(False)
    True
    >>> Null().is_valid(1)
    False
    """

    def check_value(self, value):
        if not isinstance(value, bool):
            self._failure("value should be True or False", value=value, code=codes.IS_NOT_BOOL)

    def __repr__(self):
        return "<Bool>"


class ToBool(Trafaret):
    """
    >>> extract_error(ToBool(), 'aloha')
    "value can't be converted to Bool"
    >>> ToBool().check(1)
    True
    >>> ToBool().check(0)
    False
    >>> ToBool().check('y')
    True
    >>> ToBool().check('n')
    False
    >>> ToBool().check(None)
    False
    >>> ToBool().check('1')
    True
    >>> ToBool().check('0')
    False
    >>> ToBool().check('YeS')
    True
    >>> ToBool().check('No')
    False
    >>> ToBool().check(True)
    True
    >>> ToBool().check(False)
    False
    """

    true_values = ('t', 'true', 'y', 'yes', 'on', '1', '1.0')
    false_values = ('false', 'n', 'no', 'off', '0', 'none', '0.0')
    convertable = true_values + false_values

    def check_and_return(self, value):
        _value = str(value).strip().lower()
        if _value not in self.convertable:
            self._failure(
                'value can\'t be converted to Bool',
                value=value,
                code=codes.IS_NOT_CONVERTIBLE_TO_BOOL,
            )
        return _value in ('t', 'true', 'y', 'yes', 'on', '1', '1.0')

    def __repr__(self):
        return "<ToBool>"


class Atom(Trafaret):
    """
    >>> Atom('atom').check('atom')
    'atom'
    >>> extract_error(Atom('atom'), 'molecule')
    "value is not exactly 'atom'"
    >>> Atom('atom').is_valid('atom')
    True
    >>> Atom('atom').is_valid('molecule')
    False
    """
    __slots__ = ['value']

    def __init__(self, value):
        self.value = value

    def check_value(self, value):
        if self.value != value:
            self._failure(
                "value is not exactly '%s'" % self.value,
                value=value,
                code=codes.IS_NOT_EXACTLY,
            )


class String(Trafaret):
    """
    >>> String()
    <String>
    >>> String(allow_blank=True)
    <String(blank)>
    >>> String().check("foo")
    'foo'
    >>> extract_error(String(), "")
    'blank value is not allowed'
    >>> String(allow_blank=True).check("")
    ''
    >>> extract_error(String(), 1)
    'value is not a string'
    >>> String(min_length=2, max_length=3).check('123')
    '123'
    >>> extract_error(String(min_length=2, max_length=6), '1')
    'String is shorter than 2 characters'
    >>> extract_error(String(min_length=2, max_length=6), '1234567')
    'String is longer than 6 characters'
    >>> String(min_length=2, max_length=6, allow_blank=True)
    Traceback (most recent call last):
    ...
    AssertionError: Either allow_blank or min_length should be specified, not both
    >>> String(min_length=0, max_length=6, allow_blank=True).check('123')
    '123'
    >>> String().is_valid("foo")
    True
    >>> String().is_valid("")
    False
    >>> String().is_valid(1)
    False
    """
    str_type = STR_TYPE

    TYPE_ERROR_MESSAGE = "value is not a string"
    TYPE_ERROR_CODE = codes.IS_NOT_A_STRING

    def __init__(self, allow_blank=False, min_length=None, max_length=None):
        assert not (allow_blank and min_length), \
            "Either allow_blank or min_length should be specified, not both"
        self.allow_blank = allow_blank
        self.min_length = min_length
        self.max_length = max_length

    def check_and_return(self, value):
        if not isinstance(value, self.str_type):
            self._failure(self.TYPE_ERROR_MESSAGE, value=value, code=self.TYPE_ERROR_CODE)
        if not self.allow_blank and len(value) == 0:
            self._failure("blank value is not allowed", value=value, code=codes.EMPTY_STRING)
        elif self.allow_blank and len(value) == 0:
            return value
        if self.min_length is not None and len(value) < self.min_length:
            self._failure(
                'String is shorter than %s characters' % self.min_length,
                value=value,
                code=codes.SHORT_STRING,
            )
        if self.max_length is not None and len(value) > self.max_length:
            self._failure(
                'String is longer than %s characters' % self.max_length,
                value=value,
                code=codes.LONG_STRING,
            )
        return value

    def __repr__(self):
        return "<String(blank)>" if self.allow_blank else "<String>"


class Date(Trafaret):
    """
    Checks that value is a `datetime.date` & `datetime.datetime` instances or a string
    that is convertable to `datetime.date` object.

    >>> Date()
    <Date %Y-%m-%d>
    >>> Date('%y-%m-%d')
    <Date %y-%m-%d>
    >>> Date().check(date.today())
    datetime.date(2019, 7, 25)
    >>> Date().check(datetime.now())
    datetime.datetime(2019, 10, 6, 14, 42, 52, 431348)
    >>> Date().check("2019-07-25")
    '2019, 7, 25'
    >>> Date(format='%y-%m-%d').check('00-01-01')
    '00-01-01'
    >>> extract_error(Date(), "25-07-2019")
    'value does not match format %Y-%m-%d'
    >>> extract_error(Date(), 1564077758)
    'value cannot be converted to date'
    >>> Date().is_valid(date.today())
    True
    >>> Date().is_valid(1564077758)
    False
    """

    def __init__(self, format='%Y-%m-%d'):
        self._format = format

    def _check(self, value):
        if isinstance(value, datetime):
            return value.date()
        elif isinstance(value, date):
            return value

        try:
            extracted_date = datetime.strptime(value, self._format).date()
        except ValueError:
            self._failure(
                'value does not match format %s' % self._format,
                value=value,
                code=codes.DOES_NOT_MATCH_FORMAT
            )
        except TypeError:
            self._failure(
                'value cannot be converted to date',
                value=value,
                code=codes.IS_NOT_CONVERTIBLE_TO_DATE
            )
        else:
            return extracted_date

    def check_and_return(self, value):
        self._check(value)
        return value

    def __repr__(self):
        return '<Date {}>'.format(self._format)


class ToDate(Date):
    """
    Returns instance of `datetime.date` object if value is a string or `datetime.date` & `datetime.datetime` instances.

    >>> ToDate().check(datetime.now())
    datetime.date(2019, 10, 6)
    >>> ToDate().check("2019-07-25")
    datetime.date(2019, 7, 25)
    >>> ToDate(format='%y-%m-%d').check('00-01-01')
    datetime.date(2000, 1, 1)
    """

    def check_and_return(self, data):
        return self._check(data)

    def __repr__(self):
        return '<ToDate {}>'.format(self._format)


class DateTime(Trafaret):
    """
    Checks that value is a `datetime.datetime` instance or a string that is convertable to `datetime.datetime` object.

    >>> DateTime()
    <DateTime %Y-%m-%d %H:%M:%S>
    >>> DateTime('%Y-%m-%d %H:%M')
    <DateTime %Y-%m-%d %H:%M>
    >>> DateTime().check(datetime.now())
    datetime.datetime(2019, 7, 25, 21, 45, 37, 319284)
    >>> DateTime('%Y-%m-%d %H:%M').check("2019-07-25 21:45")
    '2019-07-25 21:45'
    >>> extract_error(DateTime(), "2019-07-25")
    'value does not match format %Y-%m-%d %H:%M:%S'
    >>> extract_error(DateTime(), date.today())
    'value cannot be converted to datetime'
    >>> DateTime('%Y-%m-%d %H:%M').is_valid("2019-07-25 21:45")
    True
    >>> extract_error(DateTime(), "2019-07-25")
    False

    """

    def __init__(self, format='%Y-%m-%d %H:%M:%S'):
        self._format = format

    def _check(self, value):
        if isinstance(value, datetime):
            return value

        try:
            extracted_datetime = datetime.strptime(value, self._format)
        except ValueError:
            self._failure(
                'value does not match format %s' % self._format,
                value=value,
                code=codes.DOES_NOT_MATCH_FORMAT
            )
        except TypeError:
            self._failure(
                'value cannot be converted to datetime',
                value=value,
                code=codes.IS_NOT_CONVERTIBLE_TO_DATETIME
            )
        else:
            return extracted_datetime

    def check_and_return(self, value):
        self._check(value)
        return value

    def __repr__(self):
        return '<DateTime {}>'.format(self._format)


class ToDateTime(DateTime):
    """
    Returns instance of `datetime.datetime` object if value is a string or `datetime.datetime` instance.

    >>> DateTime('%Y-%m-%d %H:%M').check("2019-07-25 21:45")
    datetime.datetime(2019, 7, 25, 21, 45)
    """

    def check_and_return(self, value):
        return self._check(value)

    def __repr__(self):
        return '<ToDateTime {}>'.format(self._format)


class Bytes(String):
    str_type = (BYTES_TYPE,)

    TYPE_ERROR_MESSAGE = "value is not a bytes string"
    TYPE_ERROR_CODE = codes.IS_NOT_A_BYTES_STRING


class ToBytes(Trafaret):
    """Get str and try to encode it with given encoding, utf-8 by default."""
    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def check_and_return(self, value):
        if isinstance(value, BYTES_TYPE):
            return value
        elif isinstance(value, STR_TYPE):
            try:
                return value.encode(self.encoding)
            except UnicodeError:
                raise self._failure(
                    'value cannot be encoded with %s encoding' % self.encoding,
                    value=value,
                    code=codes.CANNOT_BE_ENCODED,
                )
        else:
            self._failure(
                'value is not str/bytes type',
                value=value,
                code=codes.IS_NOT_A_STRING,
            )

    def __repr__(self):
        return '<ToBytes>'


class AnyString(String):
    str_type = (BYTES_TYPE, STR_TYPE)


class FromBytes(Trafaret):
    """ Get bytes and try to decode it with given encoding, utf-8 by default.
    It can be used like ``unicode_or_koi8r = String | FromBytes(encoding='koi8r')``
    """
    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def check_and_return(self, value):
        if not isinstance(value, BYTES_TYPE):
            self._failure(
                'value is not a bytes',
                value=value,
                code=codes.IS_NOT_BYTES,
            )
        try:
            return value.decode(self.encoding)
        except UnicodeError:
            raise self._failure(
                'value cannot be decoded with %s encoding' % self.encoding,
                value=value,
                code=codes.CANNOT_BE_DECODED,
            )

    def __repr__(self):
        return "<FromBytes>"


class SquareBracketsMeta(TrafaretMeta):
    """
    Allows usage of square brackets for List initialization

    >>> List[Int]
    <List(<Int>)>
    >>> List[Int, 1:]
    <List(min_length=1 | <Int>)>
    >>> List[:10, Int]
    <List(max_length=10 | <Int>)>
    >>> List[1:10]
    Traceback (most recent call last):
    ...
    RuntimeError: Trafaret is required for List initialization
    """

    def __getitem__(self, args):
        slice_ = None
        trafaret = None
        if not isinstance(args, tuple):
            args = (args, )
        for arg in args:
            if isinstance(arg, slice):
                slice_ = arg
            elif (
                isinstance(arg, Trafaret)
                or issubclass(arg, Trafaret)
                or isinstance(arg, type)
            ):
                trafaret = arg
        if not trafaret:
            raise RuntimeError("Trafaret is required for List initialization")
        if slice_:
            return self(
                trafaret,
                min_length=slice_.start or 0,
                max_length=slice_.stop,
            )
        return self(trafaret)


@py3metafix
class Iterable(Trafaret, ListAsyncMixin):
    """
    >>> List(Int)
    <List(<Int>)>
    >>> List(Int, min_length=1)
    <List(min_length=1 | <Int>)>
    >>> List(Int, min_length=1, max_length=10)
    <List(min_length=1, max_length=10 | <Int>)>
    >>> extract_error(List(Int), 1)
    'value is not a list'
    >>> List(Int).check([1, 2, 3])
    [1, 2, 3]
    >>> List(String).check(["foo", "bar", "spam"])
    ['foo', 'bar', 'spam']
    >>> extract_error(List(Int), [1, 2, 1 + 3j])
    {2: 'value is not int'}
    >>> List(Int, min_length=1).check([1, 2, 3])
    [1, 2, 3]
    >>> extract_error(List(Int, min_length=1), [])
    'list length is less than 1'
    >>> List(Int, max_length=2).check([1, 2])
    [1, 2]
    >>> extract_error(List(Int, max_length=2), [1, 2, 3])
    'list length is greater than 2'
    >>> extract_error(List(Int), ["a"])
    {0: "value can't be converted to int"}
    >>> List(Int).is_valid([1, 2, 3])
    True
    >>> List(Int).is_valid(1)
    False
    """

    __metaclass__ = SquareBracketsMeta
    __slots__ = ['trafaret', 'min_length', 'max_length']

    def __init__(self, trafaret, min_length=0, max_length=None):
        self.trafaret = ensure_trafaret(trafaret)
        self.min_length = min_length
        self.max_length = max_length

    def check_common(self, value):
        if not isinstance(value, AbcIterable):
            self._failure(
                "value is not iterable",
                value=value,
                code=codes.IS_NOT_A_LIST,
            )
        if len(value) < self.min_length:
            self._failure(
                "list length is less than %s" % self.min_length,
                value=value,
                code=codes.TOO_SHORT,
            )
        if self.max_length is not None and len(value) > self.max_length:
            self._failure(
                "list length is greater than %s" % self.max_length,
                value=value,
                code=codes.TOO_LONG,
            )

    def transform(self, value, context=None):
        self.check_common(value)
        lst = []
        errors = {}
        for index, item in enumerate(value):
            try:
                lst.append(self.trafaret(item, context=context))
            except DataError as err:
                errors[index] = err
        if errors:
            raise self._failure(errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return lst

    def __repr__(self):
        r = "<List("
        options = []
        if self.min_length:
            options.append("min_length=%s" % self.min_length)
        if self.max_length:
            options.append("max_length=%s" % self.max_length)
        r += ", ".join(options)
        if options:
            r += " | "
        r += repr(self.trafaret)
        r += ")>"
        return r


class List(Iterable):
    def check_common(self, value):
        if not isinstance(value, list):
            self._failure(
                "value is not a list",
                value=value,
                code=codes.IS_NOT_A_LIST,
            )
        super(List, self).check_common(value)


class Tuple(Trafaret, TupleAsyncMixin):
    """
    Tuple checker can be used to check fixed tuples, like (Int, Int, String).

    >>> t = Tuple(Int, Int, String)
    >>> t.check([3, 4, '5'])
    (3, 4, '5')
    >>> extract_error(t, [3, 4, 5])
    {2: 'value is not a string'}
    >>> t
    <Tuple(<Int>, <Int>, <String>)>
    """
    __slots__ = ['trafarets', 'length']

    def __init__(self, *args):
        self.trafarets = [ensure_trafaret(t) for t in args]
        self.length = len(self.trafarets)

    def check_common(self, value):
        try:
            value = tuple(value)
        except TypeError:
            self._failure(
                'value must be convertable to tuple',
                value=value,
                code=codes.TUPLE_LIKE,
            )
        if len(value) != self.length:
            self._failure(
                'value must contain %s items' % self.length,
                value=value,
                code=codes.LOT_ELEMENTS,
            )

    def transform(self, value, context=None):
        self.check_common(value)
        result = []
        errors = {}
        for idx, (item, trafaret) in enumerate(zip(value, self.trafarets)):
            try:
                result.append(trafaret(item, context=context))
            except DataError as err:
                errors[idx] = err
        if errors:
            self._failure(errors, value=value, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return tuple(result)

    def __repr__(self):
        return '<Tuple(' + ', '.join(repr(t) for t in self.trafarets) + ')>'


class Key(KeyAsyncMixin):
    """
    Helper class for Dict.

    It gets ``name``, and provides method ``extract(data)`` that extract key value
    from data through mapping ``get`` method.
    Key `__call__` method yields ``(key name, result or DataError, [touched keys])`` triples.

    You can redefine ``get_data(data, default)`` method in subclassed ``Key`` if you want to use something other
    then ``.get(...)`` method.

    Like this for the aiohttp MultiDict::

        class MDKey(t.Key):
            def get_data(self, data, default):
                return data.getall(self.name, default)
    """
    __slots__ = ['name', 'to_name', 'default', 'optional', 'trafaret']

    def __init__(self, name, default=_empty, optional=False, to_name=None, trafaret=None):
        self.name = name
        self.to_name = to_name
        self.default = default
        self.optional = optional
        self.trafaret = ensure_trafaret(trafaret) if trafaret else Any()

    def __call__(self, data, context=None):
        if self.name in data or self.default is not _empty:
            if callable(self.default):
                default = self.default()
            else:
                default = self.default
            error = None
            try:
                result = self.trafaret(self.get_data(data, default), context=context)
            except DataError as de:
                error = de
            if error:
                yield self.name, error, (self.name,)
            else:
                yield self.get_name(), result, (self.name,)
            return

        if not self.optional:
            yield self.name, DataError(error='is required', code=codes.REQUIRED), (self.name,)

    def get_data(self, data, default):
        return data.get(self.name, default)

    def set_trafaret(self, trafaret):
        self.trafaret = ensure_trafaret(trafaret)
        return self

    def __rshift__(self, name):
        self.to_name = name
        return self

    def get_name(self):
        return self.to_name or self.name

    def __repr__(self):
        return '<%s "%s"%s %s>' % (
            self.__class__.__name__,
            self.name,
            ' to "%s"' % self.to_name if getattr(self, 'to_name', False) else '',
            self.trafaret,
        )


class Dict(Trafaret, DictAsyncMixin):
    """ Dict is a most complex trafaret that going with this library.
    The main difference from other common validation libs is that `Dict`
    trafaret does not know anything about actual keys. Every key that Dict works
    with knows itself how to get value from given mapping and if this will be one value
    or two or multi values.

    So `Dict` cannot make any assumptions about whats going on other then contract with
    a `key` implementation. And we need to note, that any callable can be `key`, not
    only `Key` subclasses.
    So we need to look at the `key` contract:

    .. code-block:: python

        key_instance(data: Mapping) -> Sequence[
            name_to_store,
            result or DataError,
            Sequence[touched keys],
        ]

    It is a bit complex, so let me explain it to you. Every key instance get this data that
    `Dict` trying to check. Then `key` will return one or multiple results and it is
    common for `key` to be generator.

    Every result is three component tuple
        1. Key for the result dict. For standard `Key` it is result key in case of successful check
        or original key name if there was an error
        2. Result if keys trafaret check was successful or DataError instance otherwise
        3. An iterable with all keys of original mapping that this `key` touched

    With this tricky interface `key` in our lib can do anything you can imagine. Like work
    with MultiDicts, compare keys, get subdicts and check them independently from main one.

    Why we need this third extra iterable with touched names? Because our Dict can check that
    all keys were consumed and what to do with extras.

    Arguments:

    * Dict accepts keys as `*args`
    * if first argument to Dict is a `dict` then its keys will be merged with args keys and
      this `dict` values must be trafarets. If key of this `dict` is a str, then Dict will create
      `Key` instance with this key as Key name and value as its trafaret. If `key` is a `Key` instance
      then Dict will call this key `set_trafaret` method.

    `allow_extra` argument can be a list of keys, or `'*'` for any, that will be checked against
    `allow_extra_trafaret` or `Any`.

    `ignore_extra` argument can be a list of keys, or `'*'` for any, that will be ignored.
    """
    __slots__ = ['extras', 'extras_trafaret', 'allow_any', 'ignore', 'ignore_any', 'keys', '_keys']

    def __init__(self, *args, **trafarets):
        if args and isinstance(args[0], AbcMapping):
            keys = args[0]
            args = args[1:]
        else:
            keys = {}
        if any(not callable(key) for key in args):
            raise RuntimeError('Keys in single attributes must be callables')

        # extra
        allow_extra = trafarets.pop('allow_extra', [])
        allow_extra_trafaret = trafarets.pop('allow_extra_trafaret', Any)
        self.extras_trafaret = ensure_trafaret(allow_extra_trafaret)
        self.allow_any = '*' in allow_extra
        self.extras = [name for name in allow_extra if name != '*']
        # ignore
        ignore_extra = trafarets.pop('ignore_extra', [])
        self.ignore_any = '*' in ignore_extra
        self.ignore = [name for name in ignore_extra if name != '*']

        self.keys = list(args)
        for key, trafaret in itertools.chain(trafarets.items(), keys.items()):
            key_ = Key(key) if isinstance(key, STR_TYPES) else key
            if not callable(key_) and not hasattr(key_, 'async_call'):
                raise RuntimeError('Non callable Keys are not supported')
            key_.set_trafaret(ensure_trafaret(trafaret))
            self.keys.append(key_)
        # optimized version without runtime check for context arg
        self._keys = []
        for key in self.keys:
            self._keys.append(with_context_caller(key))

    def _clone_args(self):
        """ return args to create new Dict clone
        """
        keys = list(self.keys)
        kw = {}
        if self.allow_any or self.extras:
            kw['allow_extra'] = list(self.extras)
            if self.allow_any:
                kw['allow_extra'].append('*')
            kw['allow_extra_trafaret'] = self.extras_trafaret
        if self.ignore_any or self.ignore:
            kw['ignore_extra'] = list(self.ignore)
            if self.ignore_any:
                kw['ignore_extra'].append('*')
        return keys, kw

    def allow_extra(self, *names, **kw):
        """ multi arguments that represents attribute names or `*`.
        Will allow unconsumed by other keys attributes for given names
        or all if includes `*`.
        Also you can pass `trafaret` keyword argument to set `Trafaret`
        instance for this extra args, or it will be `Any`.
        Method creates `Dict` clone.
        """
        keys, dictkw = self._clone_args()
        allow_extra = dictkw.setdefault('allow_extra', [])
        allow_extra.extend(names)
        if 'trafaret' in kw:
            dictkw['allow_extra_trafaret'] = kw['trafaret']
        return self.__class__(*keys, **dictkw)

    def ignore_extra(self, *names):
        """ multi arguments that represents attribute names or `*`.
        Will ignore unconsumed by other keys attribute names for given names
        or all if includes `*`.
        Method creates `Dict` clone.
        """
        keys, kw = self._clone_args()
        ignore_extra = kw.setdefault('ignore_extra', [])
        ignore_extra.extend(names)
        return self.__class__(*keys, **kw)

    def transform(self, value, context=None):
        if not isinstance(value, AbcMapping):
            self._failure(
                "value is not a dict",
                value=value,
                code=codes.IS_NOT_A_DICT,
            )
        collect = {}
        errors = {}
        touched_names = []
        for key in self._keys:
            for k, v, names in key(value, context=context):
                if isinstance(v, DataError):
                    errors[k] = v
                else:
                    collect[k] = v
                touched_names.extend(names)

        if not self.ignore_any:
            for key in value:
                if key in touched_names:
                    continue
                if key in self.ignore:
                    continue
                if not self.allow_any and key not in self.extras:
                    if key in collect:
                        errors[key] = DataError(
                            "%s key was shadowed" % key,
                            code=codes.SHADOWED,
                        )
                    else:
                        errors[key] = DataError(
                            "%s is not allowed key" % key,
                            code=codes.NOT_ALLOWED,
                        )
                elif key in collect:
                    errors[key] = DataError(
                        "%s key was shadowed" % key,
                        code=codes.SHADOWED,
                    )
                else:
                    try:
                        collect[key] = self.extras_trafaret(value[key])
                    except DataError as de:
                        errors[key] = de
        if errors:
            self._failure(error=errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return collect

    def __repr__(self):
        r = "<Dict("
        options = []
        if self.allow_any:
            options.append("any")
        if self.ignore:
            options.append("ignore=(%s)" % (", ".join(self.ignore)))
        if self.extras:
            options.append("extras=(%s)" % (", ".join(self.extras)))
        r += ", ".join(options)
        if options:
            r += " | "
        options = []
        for key in sorted(self.keys, key=lambda k: repr(k)):
            options.append(repr(key))
        r += ", ".join(options)
        r += ")>"
        return r

    def merge(self, other):
        """
        Extends one Dict with other Dict Key`s or Key`s list,
        or dict instance supposed for Dict
        """
        extra = self.extras
        if isinstance(other, Dict):
            other_keys = other.keys
            extra += other.extras
            ignore = '*' if (self.ignore_any or other.ignore_any) else self.ignore + other.ignore
        elif isinstance(other, (list, tuple)):
            other_keys = list(other)
            ignore = self.ignore
        elif isinstance(other, dict):
            return self.__class__(other, *self.keys)
        else:
            raise TypeError('You must merge Dict only with Dict'
                            ' or list of Keys')
        return self.__class__(*(self.keys + other_keys), ignore_extra=ignore,
                              allow_extra=extra)

    __add__ = merge


def DictKeys(keys):
    """
    Checks if dict has all given keys

    :param keys:
    :type keys:
    """
    req = [(Key(key), Any) for key in keys]
    return Dict(dict(req))


class Mapping(Trafaret, MappingAsyncMixin):
    """
    Mapping gets two trafarets as arguments, one for key and one for value,
    like `Mapping(t.Int, t.List(t.Str))`.
    """
    __slots__ = ['key', 'value']

    def __init__(self, key, value):
        self.key = ensure_trafaret(key)
        self.value = ensure_trafaret(value)

    def transform(self, mapping, context=None):
        if not isinstance(mapping, AbcMapping):
            self._failure(
                "value is not a dict",
                value=mapping,
                code=codes.IS_NOT_A_DICT,
            )
        checked_mapping = {}
        errors = {}
        for key, value in mapping.items():
            pair_errors = {}
            try:
                checked_key = self.key(key, context=context)
            except DataError as err:
                pair_errors['key'] = err
            try:
                checked_value = self.value(value, context=context)
            except DataError as err:
                pair_errors['value'] = err
            if pair_errors:
                errors[key] = DataError(error=pair_errors, code=codes.PAIR_MEMBERS_DID_NOT_MATCH)
            else:
                checked_mapping[checked_key] = checked_value
        if errors:
            self._failure(errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return checked_mapping

    def __repr__(self):
        return "<Mapping(%r => %r)>" % (self.key, self.value)


class Enum(Trafaret):
    """
    >>> trafaret = Enum("foo", "bar", 1) >> ignore
    >>> trafaret
    <Enum('foo', 'bar', 1)>
    >>> trafaret.check("foo")
    >>> trafaret.check(1)
    >>> extract_error(trafaret, 2)
    "value doesn't match any variant"
    >>> trafaret.is_valid(1)
    True
    >>> trafaret.is_valid(2)
    False
    """
    __slots__ = ['variants']

    def __init__(self, *variants):
        self.variants = variants[:]

    def check_value(self, value):
        if value not in self.variants:
            self._failure(
                "value doesn't match any variant",
                value=value,
                code=codes.DOES_NOT_MATCH_ANY,
            )

    def __repr__(self):
        return "<Enum(%s)>" % (", ".join(repr(v) for v in self.variants))


class Callable(Trafaret):
    """
    >>> (Callable() >> ignore).check(lambda: 1)
    >>> extract_error(Callable(), 1)
    'value is not callable'
    >>> (Callable() >> ignore).is_valid(lambda: 1)
    True
    >>> Callable().is_valid(1)
    False
    """

    def check_value(self, value):
        if not callable(value):
            self._failure(
                "value is not callable",
                value=value,
                code=codes.IS_NOT_CALLABLE,
            )

    def __repr__(self):
        return "<Callable>"


class Call(Trafaret, CallAsyncMixin):
    """
    >>> def validator(value):
    ...     if value != "foo":
    ...         return DataError("I want only foo!")
    ...     return 'foo'
    ...
    >>> trafaret = Call(validator)
    >>> trafaret
    <Call(validator)>
    >>> trafaret.check("foo")
    'foo'
    >>> extract_error(trafaret, "bar")
    'I want only foo!'
    """
    __slots__ = ['fn']

    def __init__(self, fn):
        if not callable(fn):
            raise RuntimeError("Call argument should be callable")
        args = set(get_callable_args(fn))
        self.supports_context = 'context' in args
        self.fn = fn

    def transform(self, value, context=None):
        if self.supports_context:
            res = self.fn(value, context=context)
        else:
            res = self.fn(value)
        if isinstance(res, DataError):
            raise res
        else:
            return res

    def __repr__(self):
        return "<Call(%s)>" % self.fn.__name__


class Forward(Trafaret, ForwardAsyncMixin):
    """
    >>> node = Forward()
    >>> node << Dict(name=String, children=List[node])
    >>> node
    <Forward(<Dict(children=<List(<recur>)>, name=<String>)>)>
    >>> node.check({"name": "foo", "children": []}) == {'children': [], 'name': 'foo'}
    True
    >>> extract_error(node, {"name": "foo", "children": [1]})
    {'children': {0: 'value is not a dict'}}
    >>> node.check({"name": "foo", "children": [ \
                        {"name": "bar", "children": []} \
                     ]}) == {'children': [{'children': [], 'name': 'bar'}], 'name': 'foo'}
    True
    >>> empty_node = Forward()
    >>> empty_node
    <Forward(None)>
    >>> extract_error(empty_node, 'something')
    'trafaret not set yet'
    """

    def __init__(self):
        self.trafaret = None
        self._recur_repr = False

    def __lshift__(self, trafaret):
        self.provide(trafaret)

    def provide(self, trafaret):
        if self.trafaret:
            raise RuntimeError("trafaret for Forward is already specified")
        self.trafaret = ensure_trafaret(trafaret)

    def transform(self, value, context=None):
        if self.trafaret is None:
            self._failure(
                'trafaret not set yet',
                value=value,
                code=codes.TRAFARET_IS_NOT_SET,
            )
        return self.trafaret(value, context=context)

    def __repr__(self):
        # XXX not threadsafe
        if self._recur_repr:
            return "<recur>"
        self._recur_repr = True
        r = "<Forward(%r)>" % self.trafaret
        self._recur_repr = False
        return r


class GuardError(DataError):
    """
    Raised when guarded function gets invalid arguments,
    inherits error message from corresponding DataError
    """

    pass


def guard(trafaret=None, **kwargs):
    """
    Decorator for protecting function with trafarets

    >>> @guard(a=String, b=Int, c=String)
    ... def fn(a, b, c="default"):
    ...     '''docstring'''
    ...     return (a, b, c)
    ...
    >>> fn.__module__ = None
    >>> help(fn)
    Help on function fn:
    <BLANKLINE>
    fn(*args, **kwargs)
        guarded with <Dict(a=<String>, b=<Int>, c=<String>)>
    <BLANKLINE>
        docstring
    <BLANKLINE>
    >>> fn("foo", 1)
    ('foo', 1, 'default')
    >>> extract_error(fn, "foo", 1, 2)
    {'c': 'value is not a string'}
    >>> extract_error(fn, "foo")
    {'b': 'is required'}
    >>> g = guard(Dict())
    >>> c = Forward()
    >>> c << Dict(name=str, children=List[c])
    >>> g = guard(c)
    >>> g = guard(Int())
    Traceback (most recent call last):
    ...
    RuntimeError: trafaret should be instance of Dict or Forward
    """
    if (
        trafaret
        and not isinstance(trafaret, Dict)
        and not isinstance(trafaret, Forward)
    ):
        raise RuntimeError("trafaret should be instance of Dict or Forward")
    elif trafaret and kwargs:
        raise RuntimeError("choose one way of initialization,"
                           " trafaret or kwargs")
    if not trafaret:
        trafaret = Dict(**kwargs)

    def wrapper(fn):
        argspec = getargspec(fn)

        @functools.wraps(fn)
        def decor(*args, **kwargs):
            fnargs = argspec.args
            if fnargs and fnargs[0] in ['self', 'cls']:
                obj = args[0]
                fnargs = fnargs[1:]
                checkargs = args[1:]
            else:
                obj = None
                checkargs = args

            defaults = zip(reversed(fnargs), reversed(argspec.defaults or ()))
            kw_only_defaults = (argspec.kwonlydefaults or {}).items()

            try:
                call_args = dict(
                    itertools.chain(defaults, kw_only_defaults, zip(fnargs, checkargs), kwargs.items())
                )
                converted = trafaret(call_args)
            except DataError as err:
                raise GuardError(error=err.error)
            return fn(obj, **converted) if obj else fn(**converted)
        decor.__doc__ = "guarded with %r\n\n" % trafaret + (decor.__doc__ or "")
        return decor
    return wrapper


def ignore(val):
    """
    Stub to ignore value from trafaret
    Use it like:

    >>> a = Int >> ignore
    >>> a.check(7)
    """
    pass


def catch(checker, *a, **kw):
    """
    Helper for tests - catch error and return it as dict
    """
    try:
        return checker(*a, **kw)
    except DataError as error:
        return error


catch_error = catch


def extract_error(checker, *a, **kw):
    """
    Helper for tests - catch error and return it as dict
    """

    res = catch_error(checker, *a, **kw)
    if isinstance(res, DataError):
        return res.as_dict()
    return res



IS_NOT_NULL = 'is_not_null'
IS_NOT_BOOL = 'is_not_bool'
IS_NOT_CONVERTIBLE_TO_BOOL = 'is_not_convertible_to_bool'

NOTHING_MATCH = 'nothing_match'

# String
IS_NOT_A_STRING = 'is_not_a_string'
IS_NOT_A_BYTES_STRING = 'is_not_a_bytes_string'
EMPTY_STRING = 'empty_string'
SHORT_STRING = 'short_string'
LONG_STRING = 'long_string'
# Regexp
DOES_NOT_MATCH_RE = 'does_not_match_re'

WRONG_TYPE = 'wrong_type'
IS_NOT_A_NUMBER = 'is_not_a_number'
DOES_NOT_MATCH_ANY = 'does_not_match_any'

TOO_BIG = 'too_big'
TOO_SMALL = 'too_small'

IS_NOT_INT = 'is_not_int'
INVALID_DECIMAL = 'invalid_decimal'

IS_NOT_EXACTLY = 'is_not_exactly'
TRAFARET_IS_NOT_SET = 'trafaret_is_not_set'
IS_NOT_CALLABLE = 'is_not_callable'

IS_NOT_BYTES = 'is_not_bytes'
CANNOT_BE_DECODED = 'cannot_be_decoded'
CANNOT_BE_ENCODED = 'cannot_be_encoded'

IS_NOT_A_LIST = 'is_not_a_list'
TOO_SHORT = 'too_short'
TOO_LONG = 'too_long'

TUPLE_LIKE = 'tuple_like'
LOT_ELEMENTS = 'lot_elements'

PAIR_MEMBERS_DID_NOT_MATCH = 'pair_members_did_not_match'
SOME_ELEMENTS_DID_NOT_MATCH = 'some_elements_did_not_match'

# Keys and dict
IS_NOT_A_DICT = 'is_not_a_dict'
REQUIRED = 'required'
SHADOWED = 'shadowed'
NOT_ALLOWED = 'not_allowed'
MAPPING_FAILED = 'mapping_failed'
# extra keys
MUST_BE_EQUAL = 'must_be_equal'
ONLY_ONE_MUST_BE_DEFINED = 'only_one_must_be_defined'
ONE_IS_REQUIRED = 'one_is_required'

# Internet
IS_NOT_VALID_EMAIL = 'is_not_valid_email'
IS_NOT_VALID_URL = 'is_not_valid_url'
IS_NOT_IPv4 = 'is_not_ipv4'
IS_NOT_IPv6 = 'is_not_ipv6'
IS_NOT_IP = 'is_not_ip'

# Date & DateTime
DOES_NOT_MATCH_FORMAT = 'does_not_match_format'
IS_NOT_CONVERTIBLE_TO_DATE = 'is_not_convertible_to_date'
IS_NOT_CONVERTIBLE_TO_DATETIME = 'is_not_convertible_to_datetime'


# Contrib
NOT_DATETIME = 'not_datetime'
NOT_DATE = 'not_date'



import trafaret as t
from . import codes


class KeysSubset(t.Key):
    """
    From checkers and converters dict must be returned. Some for errors.

    >>> from . import extract_error, Mapping, String
    >>> cmp_pwds = lambda x: {'pwd': x['pwd'] if x.get('pwd') == x.get('pwd1') else DataError('Not equal')}
    >>> d = Dict({KeysSubset('pwd', 'pwd1'): cmp_pwds, 'key1': String})
    >>> sorted(d.check({'pwd': 'a', 'pwd1': 'a', 'key1': 'b'}).keys())
    ['key1', 'pwd']
    >>> extract_error(d.check, {'pwd': 'a', 'pwd1': 'c', 'key1': 'b'})
    {'pwd': 'Not equal'}
    >>> extract_error(d.check, {'pwd': 'a', 'pwd1': None, 'key1': 'b'})
    {'pwd': 'Not equal'}
    >>> get_values = (lambda d, keys: [d[k] for k in keys if k in d])
    >>> join = (lambda d: {'name': ' '.join(get_values(d, ['name', 'last']))})
    >>> Dict({KeysSubset('name', 'last'): join}).check({'name': 'Adam', 'last': 'Smith'})
    {'name': 'Adam Smith'}
    """

    def __init__(self, *keys):
        self.keys = keys
        self.name = '[%s]' % ', '.join(self.keys)
        self.trafaret = t.Any()

    def __call__(self, data):
        subdict = dict((k, data.get(k)) for k in self.keys if k in data)
        keys_names = self.keys
        res = t.catch_error(self.trafaret, subdict)
        if isinstance(res, t.DataError):
            for k, e in res.error.items():
                if not isinstance(e, t.DataError):
                    raise RuntimeError('Please use DataError instance')
                yield k, e, keys_names
        else:
            for k, v in res.items():
                yield k, v, keys_names


def subdict(name, *keys, **kw):
    """
    Subdict key.

    Takes a `name`, any number of keys as args and keyword argument `trafaret`.
    Use it like:
    ::

        def check_passwords_equal(data):
            if data['password'] != data['password_confirm']:
                return t.DataError('Passwords are not equal')
            return data['password']

        passwords_key = subdict(
            'password',
            t.Key('password', trafaret=check_password),
            t.Key('password_confirm', trafaret=check_password),
            trafaret=check_passwords_equal,
        )

        signup_trafaret = t.Dict(
            t.Key('email', trafaret=t.Email),
            passwords_key,
        )

    """
    trafaret = kw.pop('trafaret')  # coz py2k

    def inner(data, context=None):
        errors = False
        preserve_output = []
        touched = set()
        collect = {}
        for key in keys:
            for k, v, names in key(data, context=context):
                touched.update(names)
                preserve_output.append((k, v, names))
                if isinstance(v, t.DataError):
                    errors = True
                else:
                    collect[k] = v
        if errors:
            for out in preserve_output:
                yield out
        elif collect:
            yield name, t.catch(trafaret, collect), touched

    return inner


def xor_key(first, second, trafaret):
    """
    xor_key - takes `first` and `second` key names and `trafaret`.

    Checks if we have only `first` or only `second` in data, not both,
    and at least one.

    Then checks key value against trafaret.
    """
    trafaret = t.ensure_trafaret(trafaret)

    def check_(value):
        if (first in value) ^ (second in value):
            key = first if first in value else second
            yield first, t.catch_error(trafaret, value[key]), (key,)
        elif first in value and second in value:
            yield (
                first,
                t.DataError(
                    error='correct only if {} is not defined'.format(second),
                    code=codes.ONLY_ONE_MUST_BE_DEFINED
                ),
                (first,)
            )
            yield (
                second,
                t.DataError(
                    error='correct only if {} is not defined'.format(first),
                    code=codes.ONLY_ONE_MUST_BE_DEFINED
                ),
                (second,)
            )
        else:
            yield (
                first,
                t.DataError(
                    error='is required if {} is not defined'.format(second),
                    code=codes.ONE_IS_REQUIRED
                ),
                (first,)
            )
            yield (
                second,
                t.DataError(
                    error='is required if {} is not defined'.format(first),
                    code=codes.ONE_IS_REQUIRED
                ),
                (second,)
            )

    return check_


def confirm_key(name, confirm_name, trafaret):
    """
    confirm_key - takes `name`, `confirm_name` and `trafaret`.

    Checks if data['name'] equals data['confirm_name'] and both
    are valid against `trafaret`.
    """
    def check_(value):
        first, second = None, None
        if name in value:
            first = value[name]
            yield name, t.catch_error(trafaret, first), (name,)
        else:
            yield name, t.DataError('is required', code=codes.REQUIRED), (name,)
        if confirm_name in value:
            second = value[confirm_name]
            yield confirm_name, t.catch_error(trafaret, second), (confirm_name,)
        else:
            yield confirm_name, t.DataError('is required', code=codes.REQUIRED), (confirm_name,)
        if not (first and second):
            return
        if first != second:
            yield (
                confirm_name,
                t.DataError(
                    'must be equal to {}'.format(name),
                    code=codes.MUST_BE_EQUAL
                ),
                (confirm_name,)
            )
    return check_



import decimal
import numbers
from .base import Trafaret, TrafaretMeta
from .lib import (
    py3metafix,
    STR_TYPES,
)
from . import codes


class NumberMeta(TrafaretMeta):
    """
    Allows slicing syntax for min and max arguments for
    number trafarets

    >>> Int[1:]
    <Int(gte=1)>
    >>> Int[1:10]
    <Int(gte=1, lte=10)>
    >>> Int[:10]
    <Int(lte=10)>
    >>> Float[1:]
    <Float(gte=1)>
    >>> Int > 3
    <Int(gt=3)>
    >>> 1 < (Float < 10)
    <Float(gt=1, lt=10)>
    >>> (Int > 5).check(10)
    10
    >>> extract_error(Int > 5, 1)
    'value should be greater than 5'
    >>> (Int < 3).check(1)
    1
    >>> extract_error(Int < 3, 3)
    'value should be less than 3'
    """

    def __getitem__(cls, slice_):
        return cls(gte=slice_.start, lte=slice_.stop)

    def __lt__(cls, lt):
        return cls(lt=lt)

    def __le__(cls, lte):
        return cls(lte=lte)

    def __gt__(cls, gt):
        return cls(gt=gt)

    def __ge__(cls, gte):
        return cls(gte=gte)


@py3metafix
class Float(Trafaret):
    """
    Tests that value is a float or a string that is convertable to float.

    >>> Float()
    <Float>
    >>> Float(gte=1)
    <Float(gte=1)>
    >>> Float(lte=10)
    <Float(lte=10)>
    >>> Float(gte=1, lte=10)
    <Float(gte=1, lte=10)>
    >>> Float().check(1.0)
    1.0
    >>> extract_error(Float(), 1 + 3j)
    'value is not float'
    >>> extract_error(Float(), 1)
    1.0
    >>> Float(gte=2).check(3.0)
    3.0
    >>> extract_error(Float(gte=2), 1.0)
    'value is less than 2'
    >>> Float(lte=10).check(5.0)
    5.0
    >>> extract_error(Float(lte=3), 5.0)
    'value is greater than 3'
    >>> Float().check("5.0")
    5.0
    """

    __metaclass__ = NumberMeta

    convertable = STR_TYPES + (numbers.Real,)
    value_type = float

    def __init__(self, gte=None, lte=None, gt=None, lt=None):
        self.gte = gte
        self.lte = lte
        self.gt = gt
        self.lt = lt

    def _converter(self, value):
        if not isinstance(value, self.convertable):
            self._failure(
                'value is not %s' % self.value_type.__name__,
                value=value,
                code=codes.WRONG_TYPE,
            )
        try:
            return self.value_type(value)
        except ValueError:
            self._failure(
                "value can't be converted to %s" % self.value_type.__name__,
                value=value,
                code=codes.IS_NOT_A_NUMBER,
            )

    def _check(self, data):
        if not isinstance(data, self.value_type):
            value = self._converter(data)
        else:
            value = data
        if self.gte is not None and value < self.gte:
            self._failure("value is less than %s" % self.gte, value=data, code=codes.TOO_SMALL)
        if self.lte is not None and value > self.lte:
            self._failure("value is greater than %s" % self.lte, value=data, code=codes.TOO_BIG)
        if self.lt is not None and value >= self.lt:
            self._failure("value should be less than %s" % self.lt, value=data, code=codes.TOO_BIG)
        if self.gt is not None and value <= self.gt:
            self._failure("value should be greater than %s" % self.gt, value=data, code=codes.TOO_SMALL)
        return value

    def check_and_return(self, data):
        self._check(data)
        return data

    def __lt__(self, lt):
        return type(self)(gte=self.gte, lte=self.lte, gt=self.gt, lt=lt)

    def __le__(self, lte):
        return type(self)(gte=self.gte, lte=lte, gt=self.gt, lt=self.lt)

    def __gt__(self, gt):
        return type(self)(gte=self.gte, lte=self.lte, gt=gt, lt=self.lt)

    def __ge__(self, gte):
        return type(self)(gte=gte, lte=self.lte, gt=self.gt, lt=self.lt)

    def __repr__(self):
        r = "<%s" % type(self).__name__
        options = []
        for param in ("gte", "lte", "gt", "lt"):
            if getattr(self, param) is not None:
                options.append("%s=%s" % (param, getattr(self, param)))
        if options:
            r += "(%s)" % (", ".join(options))
        r += ">"
        return r


class ToFloat(Float):
    """Checks that value is a float.
    Or if value is a string converts this string to float
    """
    def check_and_return(self, data):
        return self._check(data)


class Int(Float):
    """
    >>> Int()
    <Int>
    >>> Int().check(5)
    5
    >>> extract_error(Int(), 1.1)
    'value is not int'
    >>> extract_error(Int(), 1 + 1j)
    'value is not int'
    """

    value_type = int

    def _converter(self, value):
        if isinstance(value, float):
            if not value.is_integer():
                self._failure('value is not int', value=value, code=codes.IS_NOT_INT)
        return super(Int, self)._converter(value)


class ToInt(Int):
    def check_and_return(self, data):
        return self._check(data)


class ToDecimal(Float):
    value_type = decimal.Decimal

    def check_and_return(self, data):
        return self._check(data)

    def _converter(self, value):
        try:
            return self.value_type(value)
        except (ValueError, decimal.InvalidOperation):
            self._failure(
                'value can\'t be converted to Decimal',
                value=value,
                code=codes.INVALID_DECIMAL,
            )



from .lib import _empty, STR_TYPES


class DataError(ValueError):
    """
    Error with data preserve
    error can be a message or None if error raised in childs
    data can be anything
    """
    __slots__ = ['error', 'name', 'value', 'trafaret', 'code']

    error_code = 'unknown'

    def __init__(self, error=None, name=None, value=_empty, trafaret=None, code=None):
        """
        :attribute error: can be a string or a dict[string, dataerror]
        :attribute name:
        :attribute value: validated value that leads to this error
        :attribute trafaret: trafaret raised error
        :attribute code: code for error, like `value_is_too_big`
        """
        if not isinstance(error, STR_TYPES + (dict, )):
            raise RuntimeError('Only str or dict is supported, got %r' % error)
        self.error = error
        self.name = name
        self.value = value
        self.trafaret = trafaret
        self.code = code or self.__class__.error_code
        # if self.code == 'unknown':
        #     raise RuntimeError()

    def __str__(self, value=False):
        if value and self.value != _empty:
            return '%s, got %r' % (str(self.error), self.value)
        else:
            return str(self.error)

    def __repr__(self):
        return 'DataError(%r)' % str(self)

    def to_struct(self, value=False):
        if isinstance(self.error, dict):
            return {
                'code': self.code,
                'nested': dict(
                    (k, v.to_struct(value=value) if isinstance(v, DataError) else v)
                    for k, v in self.error.items()
                ),
            }
        return {
            'code': self.code,
            'message': self.__str__(value=value),
        }

    def as_dict(self, value=False):
        """Use `to_struct` if need consistency"""
        if not isinstance(self.error, dict):
            return self.__str__(value=value)

        return dict(
            (k, v.as_dict(value=value) if isinstance(v, DataError) else v)
            for k, v in self.error.items()
        )



import re
from .base import Trafaret, String
from .lib import STR_TYPES
from . import codes


class RegexpRaw(Trafaret):
    """
    Check if given string match given regexp
    """
    __slots__ = ('regexp', 'raw_regexp')

    def __init__(self, regexp, re_flags=0):
        self.regexp = re.compile(regexp, re_flags) if isinstance(regexp, STR_TYPES) else regexp
        self.raw_regexp = self.regexp.pattern if self.regexp else None

    def check_and_return(self, value):
        if not isinstance(value, STR_TYPES):
            self._failure("value is not a string", value=value, code=codes.IS_NOT_A_STRING)
        match = self.regexp.match(value)
        if not match:
            self._failure('does not match pattern %s' % self.raw_regexp, value=value, code=codes.DOES_NOT_MATCH_RE)
        return match

    def __repr__(self):
        return '<Regexp "%s">' % self.raw_regexp


class Regexp(RegexpRaw):
    def check_and_return(self, value):
        return super(Regexp, self).check_and_return(value).group()


class RegexpString(String, Regexp):
    __slots__ = ()
    # regex: str
    str_method = None

    def __init__(self, *args, **kwargs):
        String.__init__(self, *args, **kwargs)
        Regexp.__init__(self, self.regex)

    def check_and_return(self, value):
        str_value = String.check_and_return(self, value)
        if self.str_method is not None:
            str_value = getattr(str_value, self.str_method)()
        return Regexp.check_and_return(self, str_value)

    def __repr__(self):
        return '<RegexpString "%s">' % self.regex



from dateutil.parser import parse
from datetime import datetime, date
from .. import Trafaret
from ..lib import STR_TYPES
from .. import codes


class DateTime(Trafaret):
    """Class for support parsing date time in RFC3339 formats
    via dateutil.parse helper
    """
    convertable = STR_TYPES + (datetime,)
    value_type = datetime

    def __init__(self, allow_blank=False):
        self.allow_blank = allow_blank

    def __repr__(self):
        return "<DateTime(blank)>" if self.allow_blank else "<DateTime>"

    def check_and_return(self, value):
        if isinstance(value, datetime):
            return value
        try:
            return parse(value)
        except (ValueError, TypeError, OverflowError) as e:
            self._failure(str(e), code=codes.NOT_DATETIME)


class Date(Trafaret):
    """Class for support parsing dates in RFC3339 formats
    via dateutil.parse helper
    """
    convertable = STR_TYPES + (date,)
    value_type = date

    def __init__(self, allow_blank=False):
        self.allow_blank = allow_blank

    def __repr__(self):
        return "<Date(blank)>" if self.allow_blank else "<Date>"

    def check_and_return(self, value):
        if isinstance(value, datetime):
            return value.date()
        elif isinstance(value, date):
            return value
        try:
            return parse(value).date()
        except (ValueError, TypeError, OverflowError) as e:
            self._failure(str(e), code=codes.NOT_DATE)



import inspect
from .lib import AbcMapping
from .dataerror import DataError
from .lib import (
    _empty,
)
from . import codes


class TrafaretAsyncMixin:
    async def async_check(self, value, context=None):
        if hasattr(self, 'async_transform'):
            return (await self.async_transform(value, context=context))
        return self.check(value, context=context)


class OrAsyncMixin:
    async def async_transform(self, value, context=None):
        errors = []
        for trafaret in self.trafarets:
            try:
                return (await trafaret.async_check(value, context=context))
            except DataError as e:
                errors.append(e)
        self._failure(dict(enumerate(errors)), code=codes.NOTHING_MATCH)


class AndAsyncMixin:
    async def async_transform(self, value, context=None):
        res = await self.trafaret.async_check(value, context=context)
        res = await self.other.async_check(res, context=context)
        return res


class ListAsyncMixin:
    async def async_transform(self, value, context=None):
        self.check_common(value)
        lst = []
        errors = {}
        for index, item in enumerate(value):
            try:
                lst.append(await self.trafaret.async_check(item, context=context))
            except DataError as err:
                errors[index] = err
        if errors:
            self._failure(error=errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return lst


class TupleAsyncMixin:
    async def async_transform(self, value, context=None):
        self.check_common(value)
        result = []
        errors = {}
        for idx, (item, trafaret) in enumerate(zip(value, self.trafarets)):
            try:
                result.append(await trafaret.async_check(item, context=context))
            except DataError as err:
                errors[idx] = err
        if errors:
            self._failure(errors, value=value, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return tuple(result)


class MappingAsyncMixin:
    async def async_transform(self, mapping, context=None):
        if not isinstance(mapping, AbcMapping):
            self._failure("value is not a dict", value=mapping, code=codes.IS_NOT_A_DICT)
        checked_mapping = {}
        errors = {}
        for key, value in mapping.items():
            pair_errors = {}
            try:
                checked_key = await self.key.async_check(key, context=context)
            except DataError as err:
                pair_errors['key'] = err
            try:
                checked_value = await self.value.async_check(value, context=context)
            except DataError as err:
                pair_errors['value'] = err
            if pair_errors:
                errors[key] = DataError(error=pair_errors, code=codes.PAIR_MEMBERS_DID_NOT_MATCH)
            else:
                checked_mapping[checked_key] = checked_value
        if errors:
            self._failure(error=errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return checked_mapping


class CallAsyncMixin:
    async def async_transform(self, value, context=None):
        if not inspect.iscoroutinefunction(self.fn):
            return self.transform(value, context=context)
        if self.supports_context:
            res = await self.fn(value, context=context)
        else:
            res = await self.fn(value)
        if isinstance(res, DataError):
            raise res
        else:
            return res


class ForwardAsyncMixin:
    async def async_transform(self, value, context=None):
        if self.trafaret is None:
            self._failure('trafaret not set yet', value=value, code=codes.TRAFARET_IS_NOT_SET)
        return (await self.trafaret.async_check(value, context=context))


class DictAsyncMixin:
    async def async_transform(self, value, context=None):
        if not isinstance(value, AbcMapping):
            self._failure("value is not a dict", value=value, code=codes.IS_NOT_A_DICT)
        collect = {}
        errors = {}
        touched_names = []
        for key in self._keys:
            key_run = getattr(key, 'async_call', key)(
                value,
                context=context,
            )
            if inspect.isasyncgen(key_run):
                async for k, v, names in key_run:
                    if isinstance(v, DataError):
                        errors[k] = v
                    else:
                        collect[k] = v
                    touched_names.extend(names)
            else:
                for k, v, names in key_run:
                    if isinstance(v, DataError):
                        errors[k] = v
                    else:
                        collect[k] = v
                    touched_names.extend(names)

        if not self.ignore_any:
            for key in value:
                if key in touched_names:
                    continue
                if key in self.ignore:
                    continue
                if not self.allow_any and key not in self.extras:
                    if key in collect:
                        errors[key] = DataError("%s key was shadowed" % key, code=codes.SHADOWED)
                    else:
                        errors[key] = DataError("%s is not allowed key" % key, code=codes.NOT_ALLOWED)
                elif key in collect:
                    errors[key] = DataError("%s key was shadowed" % key, code=codes.SHADOWED)
                else:
                    try:
                        collect[key] = await self.extras_trafaret.async_check(value[key])
                    except DataError as de:
                        errors[key] = de
        if errors:
            self._failure(error=errors, code=codes.SOME_ELEMENTS_DID_NOT_MATCH)
        return collect


class KeyAsyncMixin:
    async def async_call(self, data, context=None):
        if self.name in data or self.default is not _empty:
            if callable(self.default):
                default = self.default()
            else:
                default = self.default
            try:
                value = await self.trafaret.async_check(self.get_data(data, default), context=context)
            except DataError as data_error:
                value = data_error
            yield (
                self.get_name(),
                value,
                (self.name,)
            )
            return

        if not self.optional:
            yield self.name, DataError(error='is required', code=codes.REQUIRED), (self.name,)
